﻿using System;
using System.Diagnostics;
using System.Reflection;
using System.Runtime.InteropServices;

namespace ConsoleApplication2
{
    // http://stackoverflow.com/questions/12656737/how-to-obtain-the-dll-list-of-a-specified-process-and-loop-through-it-to-check-i
    class Program
    {
        [Flags]
        public enum SymOpt : uint
        {
            CASE_INSENSITIVE = 0x00000001,
            UNDNAME = 0x00000002,
            DEFERRED_LOADS = 0x00000004,
            NO_CPP = 0x00000008,
            LOAD_LINES = 0x00000010,
            OMAP_FIND_NEAREST = 0x00000020,
            LOAD_ANYTHING = 0x00000040,
            IGNORE_CVREC = 0x00000080,
            NO_UNQUALIFIED_LOADS = 0x00000100,
            FAIL_CRITICAL_ERRORS = 0x00000200,
            EXACT_SYMBOLS = 0x00000400,
            ALLOW_ABSOLUTE_SYMBOLS = 0x00000800,
            IGNORE_NT_SYMPATH = 0x00001000,
            INCLUDE_32BIT_MODULES = 0x00002000,
            PUBLICS_ONLY = 0x00004000,
            NO_PUBLICS = 0x00008000,
            AUTO_PUBLICS = 0x00010000,
            NO_IMAGE_SEARCH = 0x00020000,
            SECURE = 0x00040000,
            SYMOPT_DEBUG = 0x80000000
        };

        [StructLayout(LayoutKind.Sequential)]
        public struct _IMAGEHLP_LINE64
        {
            public uint SizeOfStruct;
            public uint Key;
            public uint LineNumber;
            public IntPtr FileName;
            public ulong Address;
        };


        [DllImport("dbghelp.dll", SetLastError = true)]
        public static extern bool SymInitialize(IntPtr hProcess, string UserSearchPath, bool fInvadeProcess);


        [DllImport("dbghelp.dll", SetLastError = true)]
        public static extern uint SymSetOptions(SymOpt SymOptions);

        [DllImport("dbghelp.dll", SetLastError = true)]
        public static extern bool SymGetLineFromAddr64(IntPtr hProcess,
           ulong dwAddr, ref uint pdwDisplacement, ref _IMAGEHLP_LINE64 Line);

        static IntPtr Function()
        {
            // StackFrame st = new StackFrame(0, true);
            //Console.WriteLine(st.GetFileName() + ":" + st.GetFileLineNumber());

            MethodInfo mi = typeof(Program).GetMethod("Function", System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Static);
            return mi.MethodHandle.GetFunctionPointer();
        }

        static void Main(string[] args)
        {
            IntPtr funcAddress = Function();

            SymSetOptions(SymOpt.LOAD_LINES | SymOpt.DEFERRED_LOADS | SymOpt.SYMOPT_DEBUG);

            IntPtr processHandle = Process.GetCurrentProcess().Handle;
            int error;

            if (SymInitialize(processHandle, null, true) == false)
            {
                error = Marshal.GetLastWin32Error();
                Console.WriteLine("SymInitialize returned error : " + error);
                return;
            }

            uint dwDisplacement = 0;
            _IMAGEHLP_LINE64 line = new _IMAGEHLP_LINE64();

            ulong dwAddress = (ulong)funcAddress.ToInt64();

            if (SymGetLineFromAddr64(processHandle, dwAddress, ref dwDisplacement, ref line) == false)
            {
                error = Marshal.GetLastWin32Error();
                Console.WriteLine("SymGetLineFromAddr64 returned error : " + error);
                return;
            }

            Console.WriteLine(line.LineNumber);
        }
    }
}
